JOBAD.modules.register({

    /* JOBAD Interface  */
    info: {
        'identifier': 'odk.activecomp.demo',
        'title': 'ODK Active Document demo',
        'author': 'MMT developer team',
        'description': 'ODK active document demo',
        'version': '1.0',
        'dependencies': [],
        'hasCleanNamespace': false
    },

    /* Initialises this module */
    init: function(JOBADInstance, MMTRootURL){
        // the mmt root uri
        this.localStore.set("mmt_root", (typeof MMTRootURL === 'string')?MMTRootURL:'http://localhost:8080/');

        // the list of Context Types
        this.localStore.set("CONTEXT_TYPES", {
            'declaration': 'Declaration(s)',
            'symbol': 'Constant(s)'
        });

        // possible active computation actions -- read from the server
        this.localStore.set("ACTIONS", {});
    },

    //
    // general Utility functions
    //

    /** builds an MMT URI to be sent to the server
    * @param {string} path Path to build on the MMT side
    */
    buildMMTRequestURL : function(path) {
        return this.localStore.get("mmt_root") + path;
    },

    /** Updates the action menu from the server
    * @param {function} next function(success) that is called once the actions have been refreshed
    */
    updateActions: function(next){
        var next = (typeof next === 'function')?next:$.noop;
        var me = this;

        $.ajax({
            type: "GET",
            url: this.buildMMTRequestURL(":activecomp/actions"),
            datatype: "json",
            success: function(data){
                me.localStore.set("ACTIONS", data);
                next(true);
            }
        })
    },

    //
    // Building and extracting the context
    //

    /** Finds all context items inside of a forumula.
    * @param {jQuery} formula <math> element in which to find definiends
    */
    collectContextItems: function(formula){
        // we take the name inside of the <mi> tags
        return formula.find('mi').map(function(){
            return this.innerHTML;
        }).toArray();
    },

    /** Finds the element that gives that represents a given declaration
    * @param {jQuery} formula <math> element in which definiend is located
    * @param {string} pointer Pointer (i.e. XRef) to definiend
    */
    findDeclElement: function(formula, pointer){

      // find the possible scopes of our scope point
      var possibleScopes = formula.parents().toArray();

      // find all declarations in the Document
      // HACK HACK HACK Maybe use MMT for this
      var declarations = $("body").find("span");

      var scopePoint, scopeNeed;

      while(possibleScopes.length > 0){
        // current point in the scope
        scopePoint = possibleScopes.shift();

        // find the neccessary id
        // TODO: Just store the element itself
        if(typeof scopePoint !== 'undefined' && scopePoint.hasAttribute("id")){
          scopeNeed = '#'+scopePoint.getAttribute('id');
        } else {
          continue;
        }

        // filter the declarations by what we need
        var decls = declarations.filter(function(){
          var dScope = this.getAttribute('o:scope');
          var dFor = this.getAttribute('o:for');

          // TODO: Check that the XPath dScope resolves properly into scopePoint
          return dScope == scopeNeed && dFor == pointer;
        });

        if (decls.length > 0){
          return decls.eq(0);
        }
      }
  },
  
  /** semantically capitablises a string of words
   * @param {string} s String to capitalize
   * @returns {string}
   */
  capitalizeSemantic: function(s){
      var words = s.split(' ');
      var word;
      
      for(var i = 0; i < words.length; i++){
          word = words[i];
          words[i] = word[0].toUpperCase() + word.substring(1);
      }
      
      return words.join(' ');
  }, 

    /** Parses an element representing a declaration into a proper element of the context
    * @param {jQuery} declElem element representing declaration
    */
    declElement2Context: function(declElem){

      // HACK HACK HACK, get the type of the declaration
      var type = declElem.attr("class").substring(2);

      // HACK HACK HACK get the xref
      var xref = declElem.find("span.term").attr("xref");

      // HACK HACK HACK Take the description from the span
      var desc = declElem.find("span.term").text();
      
      // HACK HACK HACK Take the formula name
      var name = declElem.find("math mi").html();

      return {
        'type': type,
        'xref': xref,
        'desc': this.capitalizeSemantic(desc),
        'name': name
      };
    },

    /** Iterates over all items inside of a context
    * @param {function} onItem function(item, key, ctxType) that gets callled on each item
    * @param {function} onType function(section, ctxType) that gets called on each type of item
    */
    iterContext : function(items, onItem, onType){
        // cleanup input arguments
        var onType = (typeof onType === 'function')?onType:$.noop;
        var onItem = (typeof onItem === 'function')?onItem:$.noop;

        // variable that will store the current context
        var section;

        // iterate iv
        for(var ctxType in items){
            if(items.hasOwnProperty(ctxType)){
                section = items[ctxType];
                onType(section, ctxType);

                // iterate over the individual items
                for(var key in items[ctxType]){
                    onItem(section[key], key, ctxType);
                }
            }
        }
    },

    //
    // JOBAD Hooks
    //

    contextMenuEntries: function(target, JOBADInstance) {
      var me = this;

      var mathRoot = target.parent('math');

      // if we are not inside a math element, return
      if(mathRoot.length == 0){
        console.log("skipping In-Situ-Computation: Outside of a math element");
        return;
      }

      // the theory we are in (if it exists)
      var xref = mathRoot.attr("xref");
      if('string' !== typeof xref){
        console.log("skipping In-Situ-Computation: Missing 'xref' attribute");
        return;
      }

      // the keys for the context that we have to build
      var contextKeys = this.collectContextItems(mathRoot);

      // and build the context
      var context = {
        'math': $('<div>').append(mathRoot.clone()).html(),
        'items': {}
      };

      contextKeys.map(function(e, i){
        // find the element that represents the scope
        var declElement = me.findDeclElement(mathRoot, "#"+e);

        // and build a context out of it
        var contextElement = me.declElement2Context(declElement);

        // find the type of the context
        var contextType = contextElement['type'];
        delete contextElement['type'];

        // and append it to the appropriate sub-directory
        if(!(contextType in context['items'])){
          context['items'][contextType] = {};
        }
        context['items'][contextType][e] = contextElement;
      });

      return {
        'Active Computation': function(){
            me.updateActions(function(){
                me.showContextUI(context);
            });
        }
      };
    },

    // builds and show the UI for a context
    showContextUI: function(context){
      var me = this;

      var CONTEXT_TYPES = this.localStore.get("CONTEXT_TYPES");
      var ACTIONS = this.localStore.get("ACTIONS");

      // create a table and iterate over the context
      var table = $("<table>").css("width", "100%");
      this.iterContext(context.items, function(item, key, ctxType){
          table.append(
            $("<tr>").append(
              $("<td>").text(item["desc"] + " ").append(
                  $("<math>").append(
                      $("<mi>").html(item["name"])
                  )
              ),
              $("<td>").append(
                $("<input>")
                .attr("type", "text")
                .attr("name", key)
                .css("width", "100%")
                .data("context-item", item)
              )
            )
          );
      }, function(section, ctxType){
          table.append(
            $("<tr>").css("font-weight", "bold").append(
              $("<td>").attr("colspan", 2).text(CONTEXT_TYPES[ctxType])
            )
          );
      });

      // the result -- a placeholder for now
      var result = $("<div>").text("Press a button below to start the computation. Leave a field empty to find the value automatically. ");

      // build a set of actions
      var actions = $("<div>");

      // a function that is called upon executing an action
      var makeCallback = function(act){
        return function(){
          var values = {};

          // iterate over the items in the context
          table.find("input").map(function(idx, element){
              var $element = $(element);
              values[$element.data("context-item").xref] = $element.val();
          })

          me.evalContext(context, act, values, result);
        }
      };

      for(var m in ACTIONS){
        if(ACTIONS.hasOwnProperty(m)){
          actions.append(
            $("<a>").addClass("btn btn-default").text(ACTIONS[m]).on("click", makeCallback(m))
          )
        }
      }

      // build the context element
      var content = $("<div>").append(
        $.parseHTML(context.math),
        "<hr />",
        table,
        "<hr />",
        result,
        "<hr />",
        actions
      );

      // build the popup itself
      var displayDiv = $("<div>").addClass("modal hide large")
      .append(
        $("<div>").addClass("modal-header")
				.append(
					'<button type="button" class="close" data-dismiss="modal" aria-hidden="true">&times;</button>',
					'<h3>In-Situ Computation Manager</h3>'
				),
				content.addClass("modal-body"),
				$("<div>").addClass("modal-footer").append('<a href="#" class="btn btn-primary" data-dismiss="modal">Close</a>')
			).appendTo($("<div>").BS().appendTo("body"))
			.modal()
			.on("hidden", function(){
				displayDiv.parent().remove();
				JOBAD.UI.BSStyle();
			});

      // and do cleanup
      JOBAD.UI.BSStyle();
    },

    evalContext: function(context, action, values, target){

        var me = this;

        var objective = {
            "context": context,
            "values": values,
            "action": action
        };

        var items = context.items; // all of the items

        console.log(objective);

        $.ajax({
            type: "POST",
            url: this.buildMMTRequestURL(":activecomp?"),
            data: JSON.stringify(objective),
            contentType:"application/json; charset=utf-8",
            success: function(res) {
                // DEBUG: Log the resulting context
                console.log(res);

                var table = $("<table>").css("width", "100%").append(
                  $("<tr>").css("font-weight", "bold").append(
                    $("<td>").attr("colspan", 2).text("Computational Result")
                  )
                );

                //re-build the result
                me.iterContext(context.items, function(item, key, ctxType){
                    table.append(
                        $("<tr>").append(
                            $("<td>").text(item["desc"]),
                            $("<td>").append(
                                $("<input>")
                                    .attr("type", "text")
                                    .attr("readonly", "readonly")
                                    .attr("name", key)
                                    .val(res[item["xref"]])
                                    .css("width", "100%")
                            )
                        )
                    );
                });

                target.empty().append(table);
            },
            error: function(error){
                target.empty().append(error.responseText);
            }
        });
    }
});
